/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <strings.h>

#include <cstdlib>
#include <cstring>
#include <algorithm>
#include <unordered_map>
#include <vector>
#include <iterator>
#include <sstream>
#include <map>

#include "utils.h"
#include "path_manager.h"

static bool iCompare(const std::string& a, int aOffset, const std::string& b, int bOffset, int length)
{
	return static_cast<int>(a.length()) - length >= aOffset &&
		static_cast<int>(b.length()) - length >= bOffset &&
		std::equal(b.begin() + bOffset, b.begin() + bOffset + length, a.begin() + aOffset,
			[](unsigned char a, unsigned char b)
			{ return std::tolower(a) == std::tolower(b); });
}

bool isManagedAssembly(const std::string& fileName)
{
	return (iCompare(fileName, fileName.size()-4, ".dll", 0, 4) ||
			iCompare(fileName, fileName.size()-4, ".exe", 0, 4)) &&
			!isNativeImage(fileName);
}

bool isNativeImage(const std::string& fileName)
{
	return iCompare(fileName, fileName.size()-7, ".ni", 0, 3);
}

std::string readSelfPath()
{
	char buff[PATH_MAX];
	ssize_t len = ::readlink("/proc/self/exe", buff, sizeof(buff)-1);
	if (len != -1) {
		buff[len] = '\0';
		return std::string(buff);
	}

	return "";
}

std::string concatPath(const std::string& path1, const std::string& path2)
{
	std::string path(path1);
	if (path.back() == PATH_SEPARATOR) {
		path.append(path2);
	} else {
		path += PATH_SEPARATOR;
		path.append(path2);
	}

	return path;
}

void splitPath(const std::string& path, std::vector<std::string>& out)
{
	std::istringstream ss(path);
	std::string token;

	while (std::getline(ss, token, ':')) {
		out.push_back(token);
	}
}

std::string absolutePath(const std::string& path)
{
	std::string absPath;
	char realPath[PATH_MAX];
	if (realpath(path.c_str(), realPath) != nullptr && realPath[0] != '\0')
		absPath.assign(realPath);

	return absPath;
}

std::string baseName(const std::string& path)
{
	auto pos = path.find_last_of(PATH_SEPARATOR);
	if (pos != std::string::npos)
		return path.substr(0, pos);
	else
		return std::string(".");
	return path;
}

bool isFileExist(const std::string& path)
{
	struct stat sb;
	return stat(path.c_str(), &sb) == 0;
}

std::string stripNiDLL(const std::string& path)
{
	std::string niPath(path);
	if (path.size() < 5) return niPath;
	if (!strncasecmp(path.c_str() + path.size() - 4, ".dll", 4))
		niPath = path.substr(0, path.size()-4);
	else if (!strncasecmp(path.c_str() + path.size() - 4, ".exe", 4))
		niPath = path.substr(0, path.size()-4);

	if (!strncasecmp(niPath.c_str() + niPath.size() - 3, ".ni", 3))
		return niPath.substr(0, niPath.size()-3);

	return niPath;
}

void assembliesInDirectory(const std::vector<std::string>& directories, std::string& tpaList)
{
	std::map<std::string, std::string> assemblyList;
	std::map<std::string, std::string> tmpList;

	auto reader = [&assemblyList, &tmpList] (const std::string& path, const char* name) {
		if (isManagedAssembly(path) || isNativeImage(path)) {
			std::string dllName = stripNiDLL(name);
			std::pair<std::map<std::string, std::string>::iterator, bool> ret;
			ret = tmpList.insert(std::pair<std::string, std::string>(dllName, path));
			if (ret.second == false) {
				if (isNativeImage(path))
					tmpList[dllName] = path;
			}
		}
	};

	for (auto directory : directories) {
		scanFilesInDir(directory.c_str(), reader, 1);
		// merge scaned dll list to tpa list.
		// if the dll is already exist in the list, that is skipped.
		assemblyList.insert(tmpList.begin(), tmpList.end());
	}

	std::map<std::string, std::string>::iterator it;
	for (it = assemblyList.begin(); it != assemblyList.end(); it++)
		tpaList += it->second + ':';

	if (tpaList.back() == ':')
		tpaList.pop_back();
}

void scanFilesInDir(const std::string& directory, FileReader reader, unsigned int depth)
{
	DIR *dir;
	struct dirent* entry;
	bool isDir;

	dir = opendir(directory.c_str());

	if (dir == nullptr)
		return;

	std::vector<std::string> innerDirectories;

	while ((entry = readdir(dir)) != nullptr) {
		isDir = false;
		std::string path = concatPath(directory, entry->d_name);
		switch (entry->d_type) {
			case DT_REG: break;
			case DT_DIR:
				isDir = true;
				break;
			case DT_LNK:
			case DT_UNKNOWN:
				struct stat sb;
				if (stat(path.c_str(), &sb) == -1)
					continue;

				if (S_ISREG(sb.st_mode) || S_ISDIR(sb.st_mode))
					break;
			default:
				continue;
		}
		if (!isDir)
			reader(path, entry->d_name);
		else if (depth > 1 && strcmp(entry->d_name, ".") && strcmp(entry->d_name, ".."))
			innerDirectories.push_back(path);
	}

	if (depth != 0)
		for (auto& d : innerDirectories)
			scanFilesInDir(d.c_str(), reader, depth-1);

	closedir(dir);
}

